@@ -139,7 +139,7 @@ rmjob(void)
 			assasinated = kill(cur_daemon, SIGINT) == 0;
 			PRIV_END;
 			if (!assasinated)
-				fatal("cannot kill printer daemon");
+				fatal("cannot kill printer daemon (pid: %d)", (int)cur_daemon);
 		}
 		/*
 		 * process the files
@@ -179,7 +179,7 @@ lockchk(char *s)
 		else
 			return(0);
 	}
-	if (!getline(fp)) {
+	if (!getcfline(fp)) {
 		(void)fclose(fp);
 		return(0);		/* no daemon present */
 	}
@@ -219,7 +219,7 @@ process(char *file)
 			close(fd);
 		fatal("cannot open %s", file);
 	}
-	while (getline(cfp)) {
+	while (getcfline(cfp)) {
 		switch (line[0]) {
 		case 'U':  /* unlink associated files */
 			if (strchr(line+1, '/') || strncmp(line+1, "df", 2))
@@ -274,7 +274,7 @@ chk(char *file)
 			close(fd);
 		return(0);
 	}
-	while (getline(cfp)) {
+	while (getcfline(cfp)) {
 		if (line[0] == 'P')
 			break;
 	}
@@ -401,7 +401,7 @@ alarmer(int s)
  * Return 1 if the filename begins with 'cf'
  */
 int
-iscf(struct dirent *d)
+iscf(const struct dirent *d)
 {
 	return(d->d_name[0] == 'c' && d->d_name[1] == 'f');
 }
