@@ -1,4 +1,5 @@
-/*	$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $	*/
+/*	$OpenBSD: cmds.c,v 1.18 2003/06/02 23:36:53 millert Exp $	*/
+/*	$NetBSD: cmds.c,v 1.12 1997/10/05 15:12:06 mrg Exp $	*/
 
 /*
  * Copyright (c) 1983, 1993
@@ -13,11 +14,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *    must display the following acknowledgement:
- *	This product includes software developed by the University of
- *	California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *    may be used to endorse or promote products derived from this software
  *    without specific prior written permission.
  *
@@ -35,16 +32,16 @@
  */
 
 #ifndef lint
-static char copyright[] =
+static const char copyright[] =
 "@(#) Copyright (c) 1983, 1993\n\
 	The Regents of the University of California.  All rights reserved.\n";
 #endif /* not lint */
 
 #ifndef lint
 #if 0
-static char sccsid[] = "@(#)cmds.c	8.2 (Berkeley) 4/28/95";
+static const char sccsid[] = "@(#)cmds.c	8.2 (Berkeley) 4/28/95";
 #else
-static char rcsid[] = "$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $";
+static const char rcsid[] = "$OpenBSD: cmds.c,v 1.18 2003/06/02 23:36:53 millert Exp $";
 #endif
 #endif /* not lint */
 
@@ -52,6 +49,7 @@ static char rcsid[] = "$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $";
  * lpc -- line printer control program -- commands:
  */
 
+#include <fcntl.h>
 #include <sys/param.h>
 #include <sys/time.h>
 #include <sys/stat.h>
@@ -72,46 +70,42 @@ static char rcsid[] = "$OpenBSD: cmds.c,v 1.9 1997/07/25 18:57:24 grr Exp $";
 #include "extern.h"
 #include "pathnames.h"
 
-extern uid_t	uid, euid;
-
-static void	abortpr __P((int));
-static void	cleanpr __P((void));
-static void	disablepr __P((void));
-static int	doarg __P((char *));
-static int	doselect __P((struct dirent *));
-static void	enablepr __P((void));
-static void	prstat __P((void));
-static void	putmsg __P((int, char **));
-static int	sortq __P((const void *, const void *));
-static void	startpr __P((int));
-static void	stoppr __P((void));
-static int	touch __P((struct queue *));
-static void	unlinkf __P((char *));
-static void	upstat __P((char *));
+static void	abortpr(int);
+static void	cleanpr(void);
+static void	disablepr(void);
+static int	doarg(char *);
+static int	doselect(const struct dirent *);
+static void	enablepr(void);
+static void	prstat(void);
+static void	putmsg(int, char **);
+static int	sortq(const void *, const void *);
+static void	startpr(int);
+static void	stoppr(void);
+static int	touch(struct queue *);
+static void	unlinkf(char *);
+static void	upstat(char *);
 
 /*
  * kill an existing daemon and disable printing.
  */
 void
-doabort(argc, argv)
-	int argc;
-	char *argv[];
+doabort(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: abort {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			abortpr(1);
@@ -133,10 +127,9 @@ doabort(argc, argv)
 }
 
 static void
-abortpr(dis)
-	int dis;
+abortpr(int dis)
 {
-	register FILE *fp;
+	FILE *fp;
 	struct stat stbuf;
 	int pid, fd;
 
@@ -144,26 +137,29 @@ abortpr(dis)
 		SD = _PATH_DEFSPOOL;
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
+	PRIV_START;
 	/*
 	 * Turn on the owner execute bit of the lock file to disable printing.
 	 */
 	if (dis) {
-		seteuid(euid);
 		if (stat(line, &stbuf) >= 0) {
-			if (chmod(line, (stbuf.st_mode & 0777) | 0100) < 0)
+			stbuf.st_mode |= S_IXUSR;
+			if (chmod(line, stbuf.st_mode & 0777) < 0)
 				printf("\tcannot disable printing\n");
 			else {
 				upstat("printing disabled\n");
 				printf("\tprinting disabled\n");
 			}
 		} else if (errno == ENOENT) {
-			if ((fd = open(line, O_WRONLY|O_CREAT, 0760)) < 0)
+			if ((fd = safe_open(line, O_WRONLY|O_CREAT|O_NOFOLLOW,
+			    0760)) < 0)
 				printf("\tcannot create lock file\n");
 			else {
-				(void) close(fd);
+				(void)fchown(fd, DEFNUID, -1);
+				(void)close(fd);
 				upstat("printing disabled\n");
 				printf("\tprinting disabled\n");
 				printf("\tno daemon to abort\n");
@@ -177,16 +173,19 @@ abortpr(dis)
 	/*
 	 * Kill the current daemon to stop printing now.
 	 */
-	if ((fp = fopen(line, "r")) == NULL) {
+	fd = safe_open(line, O_RDONLY|O_NOFOLLOW, 0);
+	if (fd < 0 || (fp = fdopen(fd, "r")) == NULL) {
+		if (fd >= 0)
+			close(fd);
 		printf("\tcannot open lock file\n");
 		goto out;
 	}
-	if (!getline(fp) || flock(fileno(fp), LOCK_SH|LOCK_NB) == 0) {
-		(void) fclose(fp);	/* unlocks as well */
+	if (!getcfline(fp) || flock(fileno(fp), LOCK_SH|LOCK_NB) == 0) {
+		(void)fclose(fp);	/* unlocks as well */
 		printf("\tno daemon to abort\n");
 		goto out;
 	}
-	(void) fclose(fp);
+	(void)fclose(fp);
 	if (kill(pid = atoi(line), SIGTERM) < 0) {
 		if (errno == ESRCH)
 			printf("\tno daemon to abort\n");
@@ -195,59 +194,56 @@ abortpr(dis)
 	} else
 		printf("\tdaemon (pid %d) killed\n", pid);
 out:
-	seteuid(uid);
+	PRIV_END;
 }
 
 /*
- * Write a message into the status file.
+ * Write a message into the status file (assumes PRIV_START already called)
  */
 static void
-upstat(msg)
-	char *msg;
+upstat(char *msg)
 {
-	register int fd;
+	int fd;
 	char statfile[MAXPATHLEN];
 
 	if (cgetstr(bp, "st", &ST) == -1)
 		ST = DEFSTAT;
-	(void) snprintf(statfile, sizeof(statfile), "%s/%s", SD, ST);
-	umask(0);
-	fd = open(statfile, O_WRONLY|O_CREAT, 0664);
+	(void)snprintf(statfile, sizeof(statfile), "%s/%s", SD, ST);
+	fd = safe_open(statfile, O_WRONLY|O_CREAT|O_NOFOLLOW, 0660);
 	if (fd < 0 || flock(fd, LOCK_EX) < 0) {
 		printf("\tcannot create status file\n");
 		return;
 	}
-	(void) ftruncate(fd, 0);
+	(void)fchown(fd, DEFNUID, -1);
+	(void)ftruncate(fd, 0);
 	if (msg == (char *)NULL)
-		(void) write(fd, "\n", 1);
+		(void)write(fd, "\n", 1);
 	else
-		(void) write(fd, msg, strlen(msg));
-	(void) close(fd);
+		(void)write(fd, msg, strlen(msg));
+	(void)close(fd);
 }
 
 /*
  * Remove all spool files and temporaries from the spooling area.
  */
 void
-clean(argc, argv)
-	int argc;
-	char *argv[];
+clean(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: clean {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			cleanpr();
@@ -270,8 +266,7 @@ clean(argc, argv)
 }
 
 static int
-doselect(d)
-	struct dirent *d;
+doselect(const struct dirent *d)
 {
 	int c = d->d_name[0];
 
@@ -285,15 +280,14 @@ doselect(d)
  * by `cf', `tf', or `df', then by the sequence letter A-Z, a-z.
  */
 static int
-sortq(a, b)
-	const void *a, *b;
+sortq(const void *a, const void *b)
 {
-	struct dirent **d1, **d2;
+	const struct dirent **d1, **d2;
 	int c1, c2;
 
-	d1 = (struct dirent **)a;
-	d2 = (struct dirent **)b;
-	if ((c1 = strcmp((*d1)->d_name + 3, (*d2)->d_name + 3)))
+	d1 = (const struct dirent **)a;
+	d2 = (const struct dirent **)b;
+	if ((c1 = strcmp((*d1)->d_name + 3, (*d2)->d_name + 3)) != 0)
 		return(c1);
 	c1 = (*d1)->d_name[0];
 	c2 = (*d2)->d_name[0];
@@ -310,10 +304,10 @@ sortq(a, b)
  * Remove incomplete jobs from spooling area.
  */
 static void
-cleanpr()
+cleanpr(void)
 {
-	register int i, n;
-	register char *cp, *cp1, *lp;
+	int i, n;
+	char *cp, *cp1, *lp;
 	struct dirent **queue;
 	int nitems;
 
@@ -321,13 +315,19 @@ cleanpr()
 		SD = _PATH_DEFSPOOL;
 	printf("%s:\n", printer);
 
-	for (lp = line, cp = SD; (lp - line) < sizeof(line) && (*lp++ = *cp++);)
+	/* XXX depends on SD being non-NUL */
+	for (lp = line, cp = SD; (lp - line) < sizeof(line) &&
+	    (*lp++ = *cp++) != '\0'; )
 		;
 	lp[-1] = '/';
+	if (lp - line >= sizeof(line)) {
+		printf("\tspool directory name too long\n");
+		return;
+	}
 
-	seteuid(euid);
+	PRIV_START;
 	nitems = scandir(SD, &queue, doselect, sortq);
-	seteuid(uid);
+	PRIV_END;
 	if (nitems < 0) {
 		printf("\tcannot examine spool directory\n");
 		return;
@@ -347,9 +347,11 @@ cleanpr()
 				n++;
 			}
 			if (n == 0) {
-				strncpy(lp, cp, sizeof(line) - strlen(line) - 1);
-				line[sizeof(line) - 1] = '\0';
-				unlinkf(line);
+				if (strlcpy(lp, cp, sizeof(line) - (lp - line))
+				    >= sizeof(line) - (lp - line))
+					printf("\tpath too long, %s/%s", SD, cp);
+				else
+					unlinkf(line);
 			}
 		} else {
 			/*
@@ -357,48 +359,47 @@ cleanpr()
 			 * been skipped above) or a tf file (which can always
 			 * be removed).
 			 */
-			strncpy(lp, cp, sizeof(line) - strlen(line) - 1);
-			line[sizeof(line) - 1] = '\0';
-			unlinkf(line);
+			if (strlcpy(lp, cp, sizeof(line) - (lp - line)) >=
+			    sizeof(line) - (lp - line))
+				printf("\tpath too long, %s/%s", SD, cp);
+			else
+				unlinkf(line);
 		}
      	} while (++i < nitems);
 }
  
 static void
-unlinkf(name)
-	char	*name;
+unlinkf(char *name)
 {
-	seteuid(euid);
+	PRIV_START;
 	if (unlink(name) < 0)
 		printf("\tcannot remove %s\n", name);
 	else
 		printf("\tremoved %s\n", name);
-	seteuid(uid);
+	PRIV_END;
 }
 
 /*
  * Enable queuing to the printer (allow lpr's).
  */
 void
-enable(argc, argv)
-	int argc;
-	char *argv[];
+enable(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: enable {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			enablepr();
@@ -421,7 +422,7 @@ enable(argc, argv)
 }
 
 static void
-enablepr()
+enablepr(void)
 {
 	struct stat stbuf;
 
@@ -429,45 +430,44 @@ enablepr()
 		SD = _PATH_DEFSPOOL;
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn off the group execute bit of the lock file to enable queuing.
 	 */
-	seteuid(euid);
+	PRIV_START;
 	if (stat(line, &stbuf) >= 0) {
-		if (chmod(line, stbuf.st_mode & 0767) < 0)
+		stbuf.st_mode &= ~S_IXGRP;
+		if (chmod(line, stbuf.st_mode & 0777) < 0)
 			printf("\tcannot enable queuing\n");
 		else
 			printf("\tqueuing enabled\n");
 	}
-	seteuid(uid);
+	PRIV_END;
 }
 
 /*
  * Disable queuing.
  */
 void
-disable(argc, argv)
-	int argc;
-	char *argv[];
+disable(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: disable {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			disablepr();
@@ -490,36 +490,38 @@ disable(argc, argv)
 }
 
 static void
-disablepr()
+disablepr(void)
 {
-	register int fd;
+	int fd;
 	struct stat stbuf;
 
 	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 	/*
 	 * Turn on the group execute bit of the lock file to disable queuing.
 	 */
-	seteuid(euid);
+	PRIV_START;
 	if (stat(line, &stbuf) >= 0) {
-		if (chmod(line, (stbuf.st_mode & 0777) | 010) < 0)
+		stbuf.st_mode |= S_IXGRP;
+		if (chmod(line, stbuf.st_mode & 0777) < 0)
 			printf("\tcannot disable queuing\n");
 		else
 			printf("\tqueuing disabled\n");
 	} else if (errno == ENOENT) {
-		if ((fd = open(line, O_WRONLY|O_CREAT, 0670)) < 0)
+		if ((fd = safe_open(line, O_WRONLY|O_CREAT|O_NOFOLLOW, 0670)) < 0)
 			printf("\tcannot create lock file\n");
 		else {
-			(void) close(fd);
+			(void)fchown(fd, DEFNUID, -1);
+			(void)close(fd);
 			printf("\tqueuing disabled\n");
 		}
 	} else
 		printf("\tcannot stat lock file\n");
-	seteuid(uid);
+	PRIV_END;
 }
 
 /*
@@ -527,25 +529,23 @@ disablepr()
  * (reason for being down).
  */
 void
-down(argc, argv)
-	int argc;
-	char *argv[];
+down(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: down {all | printer} [message ...]\n");
 		return;
 	}
-	if (!strcmp(argv[1], "all")) {
+	if (strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			putmsg(argc - 2, argv + 2);
@@ -566,12 +566,10 @@ down(argc, argv)
 }
 
 static void
-putmsg(argc, argv)
-	int argc;
-	char **argv;
+putmsg(int argc, char **argv)
 {
-	register int fd;
-	register char *cp1, *cp2;
+	int fd;
+	char *cp1, *cp2;
 	char buf[1024];
 	struct stat stbuf;
 
@@ -586,61 +584,62 @@ putmsg(argc, argv)
 	 * Turn on the group execute bit of the lock file to disable queuing and
 	 * turn on the owner execute bit of the lock file to disable printing.
 	 */
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
-	seteuid(euid);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	PRIV_START;
 	if (stat(line, &stbuf) >= 0) {
-		if (chmod(line, (stbuf.st_mode & 0777) | 0110) < 0)
+		stbuf.st_mode |= (S_IXGRP|S_IXUSR);
+		if (chmod(line, stbuf.st_mode & 0777) < 0)
 			printf("\tcannot disable queuing\n");
 		else
 			printf("\tprinter and queuing disabled\n");
 	} else if (errno == ENOENT) {
-		if ((fd = open(line, O_WRONLY|O_CREAT, 0770)) < 0)
+		if ((fd = safe_open(line, O_WRONLY|O_CREAT|O_NOFOLLOW, 0770)) < 0)
 			printf("\tcannot create lock file\n");
 		else {
-			(void) close(fd);
+			(void)fchown(fd, DEFNUID, -1);
+			(void)close(fd);
 			printf("\tprinter and queuing disabled\n");
 		}
-		seteuid(uid);
+		PRIV_END;
 		return;
 	} else
 		printf("\tcannot stat lock file\n");
 	/*
 	 * Write the message into the status file.
 	 */
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, ST);
-	fd = open(line, O_WRONLY|O_CREAT, 0664);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, ST);
+	fd = safe_open(line, O_WRONLY|O_CREAT|O_NOFOLLOW, 0660);
 	if (fd < 0 || flock(fd, LOCK_EX) < 0) {
 		printf("\tcannot create status file\n");
-		seteuid(uid);
+		PRIV_END;
 		return;
 	}
-	seteuid(uid);
-	(void) ftruncate(fd, 0);
+	PRIV_END;
+	(void)fchown(fd, DEFNUID, -1);
+	(void)ftruncate(fd, 0);
 	if (argc <= 0) {
-		(void) write(fd, "\n", 1);
-		(void) close(fd);
+		(void)write(fd, "\n", 1);
+		(void)close(fd);
 		return;
 	}
 	cp1 = buf;
 	while (--argc >= 0) {
 		cp2 = *argv++;
-		while ((cp1 - buf) < sizeof(buf) && (*cp1++ = *cp2++))
+		while ((cp1 - buf) < sizeof(buf) - 1 && (*cp1++ = *cp2++))
 			;
 		cp1[-1] = ' ';
 	}
 	cp1[-1] = '\n';
 	*cp1 = '\0';
-	(void) write(fd, buf, strlen(buf));
-	(void) close(fd);
+	(void)write(fd, buf, strlen(buf));
+	(void)close(fd);
 }
 
 /*
  * Exit lpc
  */
 void
-quit(argc, argv)
-	int argc;
-	char *argv[];
+quit(int argc, char **argv)
 {
 	exit(0);
 }
@@ -649,25 +648,23 @@ quit(argc, argv)
  * Kill and restart the daemon.
  */
 void
-restart(argc, argv)
-	int argc;
-	char *argv[];
+restart(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: restart {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			abortpr(0);
@@ -695,25 +692,23 @@ restart(argc, argv)
  * Enable printing on the specified printer and startup the daemon.
  */
 void
-startcmd(argc, argv)
-	int argc;
-	char *argv[];
+startcmd(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: start {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			startpr(1);
@@ -736,8 +731,7 @@ startcmd(argc, argv)
 }
 
 static void
-startpr(enable)
-	int enable;
+startpr(int enable)
 {
 	struct stat stbuf;
 
@@ -745,45 +739,48 @@ startpr(enable)
 		SD = _PATH_DEFSPOOL;
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn off the owner execute bit of the lock file to enable printing.
+	 * If we are marking the printer "up" also turn off group execute bit.
 	 */
-	seteuid(euid);
+	PRIV_START;
 	if (enable && stat(line, &stbuf) >= 0) {
-		if (chmod(line, stbuf.st_mode & (enable==2 ? 0666 : 0677)) < 0)
+		if (enable == 2)
+			stbuf.st_mode &= ~(S_IXUSR|S_IXGRP);
+		else
+			stbuf.st_mode &= ~S_IXUSR;
+		if (chmod(line, stbuf.st_mode & 0777) < 0)
 			printf("\tcannot enable printing\n");
 		else
 			printf("\tprinting enabled\n");
 	}
+	PRIV_END;
 	if (!startdaemon(printer))
 		printf("\tcouldn't start daemon\n");
 	else
 		printf("\tdaemon started\n");
-	seteuid(uid);
 }
 
 /*
  * Print the status of each queue listed or all the queues.
  */
 void
-status(argc, argv)
-	int argc;
-	char *argv[];
+status(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
-	if (argc == 1 || argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 1 || (argc == 2 && strcmp(argv[1], "all") == 0)) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			prstat();
@@ -809,11 +806,11 @@ status(argc, argv)
  * Print the status of the printer queue.
  */
 static void
-prstat()
+prstat(void)
 {
 	struct stat stbuf;
-	register int fd, i;
-	register struct dirent *dp;
+	int fd, i;
+	struct dirent *dp;
 	DIR *dirp;
 
 	if (cgetstr(bp, "sd", &SD) == -1)
@@ -823,8 +820,11 @@ prstat()
 	if (cgetstr(bp, "st", &ST) == -1)
 		ST = DEFSTAT;
 	printf("%s:\n", printer);
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
-	if (stat(line, &stbuf) >= 0) {
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	PRIV_START;
+	i = stat(line, &stbuf);
+	PRIV_END;
+	if (i >= 0) {
 		printf("\tqueuing is %s\n",
 			(stbuf.st_mode & 010) ? "disabled" : "enabled");
 		printf("\tprinting is %s\n",
@@ -833,7 +833,10 @@ prstat()
 		printf("\tqueuing is enabled\n");
 		printf("\tprinting is enabled\n");
 	}
-	if ((dirp = opendir(SD)) == NULL) {
+	PRIV_START;
+	dirp = opendir(SD);
+	PRIV_END;
+	if (dirp == NULL) {
 		printf("\tcannot examine spool directory\n");
 		return;
 	}
@@ -849,21 +852,27 @@ prstat()
 		printf("\t1 entry in spool area\n");
 	else
 		printf("\t%d entries in spool area\n", i);
-	fd = open(line, O_RDONLY);
+	PRIV_START;
+	fd = safe_open(line, O_RDONLY|O_NOFOLLOW, 0);
+	PRIV_END;
 	if (fd < 0 || flock(fd, LOCK_SH|LOCK_NB) == 0) {
-		(void) close(fd);	/* unlocks as well */
+		(void)close(fd);	/* unlocks as well */
 		printf("\tprinter idle\n");
 		return;
 	}
-	(void) close(fd);
-	putchar('\t');
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, ST);
-	fd = open(line, O_RDONLY);
+	(void)close(fd);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, ST);
+	PRIV_START;
+	fd = safe_open(line, O_RDONLY|O_NOFOLLOW, 0);
+	PRIV_END;
 	if (fd >= 0) {
-		(void) flock(fd, LOCK_SH);
-		while ((i = read(fd, line, sizeof(line))) > 0)
-			(void) fwrite(line, 1, i, stdout);
-		(void) close(fd);	/* unlocks as well */
+		(void)flock(fd, LOCK_SH);
+		if (fstat(fd, &stbuf) == 0 && stbuf.st_size > 0) {
+			putchar('\t');
+			while ((i = read(fd, line, sizeof(line))) > 0)
+				(void)fwrite(line, 1, i, stdout);
+		}
+		(void)close(fd);	/* unlocks as well */
 	}
 }
 
@@ -872,25 +881,23 @@ prstat()
  * printing.
  */
 void
-stop(argc, argv)
-	int argc;
-	char *argv[];
+stop(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: stop {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			stoppr();
@@ -913,40 +920,42 @@ stop(argc, argv)
 }
 
 static void
-stoppr()
+stoppr(void)
 {
-	register int fd;
+	int fd;
 	struct stat stbuf;
 
 	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
 	if (cgetstr(bp, "lo", &LO) == -1)
 		LO = DEFLOCK;
-	(void) snprintf(line, sizeof(line), "%s/%s", SD, LO);
+	(void)snprintf(line, sizeof(line), "%s/%s", SD, LO);
 	printf("%s:\n", printer);
 
 	/*
 	 * Turn on the owner execute bit of the lock file to disable printing.
 	 */
-	seteuid(euid);
+	PRIV_START;
 	if (stat(line, &stbuf) >= 0) {
-		if (chmod(line, (stbuf.st_mode & 0777) | 0100) < 0)
+		stbuf.st_mode |= S_IXUSR;
+		if (chmod(line, stbuf.st_mode & 0777) < 0)
 			printf("\tcannot disable printing\n");
 		else {
 			upstat("printing disabled\n");
 			printf("\tprinting disabled\n");
 		}
 	} else if (errno == ENOENT) {
-		if ((fd = open(line, O_WRONLY|O_CREAT, 0760)) < 0)
+		if ((fd = safe_open(line, O_WRONLY|O_CREAT|O_NOFOLLOW, 0760)) < 0)
 			printf("\tcannot create lock file\n");
 		else {
-			(void) close(fd);
+			(void)fchown(fd, DEFNUID, -1);
+			(void)close(fd);
 			upstat("printing disabled\n");
 			printf("\tprinting disabled\n");
 		}
 	} else
 		printf("\tcannot stat lock file\n");
-	seteuid(uid);
+	PRIV_END;
 }
 
 struct	queue **queue;
@@ -957,11 +966,9 @@ time_t	mtime;
  * Put the specified jobs at the top of printer queue.
  */
 void
-topq(argc, argv)
-	int argc;
-	char *argv[];
+topq(int argc, char **argv)
 {
-	register int i;
+	int i;
 	struct stat stbuf;
 	int status, changed;
 
@@ -988,12 +995,12 @@ topq(argc, argv)
 		LO = DEFLOCK;
 	printf("%s:\n", printer);
 
-	seteuid(euid);
+	PRIV_START;
 	if (chdir(SD) < 0) {
 		printf("\tcannot chdir to %s\n", SD);
 		goto out;
 	}
-	seteuid(uid);
+	PRIV_END;
 	nitems = getq(&queue);
 	if (nitems == 0)
 		return;
@@ -1017,12 +1024,14 @@ topq(argc, argv)
 	 * Turn on the public execute bit of the lock file to
 	 * get lpd to rebuild the queue after the current job.
 	 */
-	seteuid(euid);
-	if (changed && stat(LO, &stbuf) >= 0)
-		(void) chmod(LO, (stbuf.st_mode & 0777) | 01);
+	PRIV_START;
+	if (changed && stat(LO, &stbuf) >= 0) {
+		stbuf.st_mode |= S_IXOTH;
+		(void)chmod(LO, stbuf.st_mode & 0777);
+	}
 
 out:
-	seteuid(uid);
+	PRIV_END;
 } 
 
 /*
@@ -1030,17 +1039,16 @@ out:
  * the control file.
  */
 static int
-touch(q)
-	struct queue *q;
+touch(struct queue *q)
 {
 	struct timeval tvp[2];
 	int ret;
 
 	tvp[0].tv_sec = tvp[1].tv_sec = --mtime;
 	tvp[0].tv_usec = tvp[1].tv_usec = 0;
-	seteuid(euid);
+	PRIV_START;
 	ret = utimes(q->q_name, tvp);
-	seteuid(uid);
+	PRIV_END;
 	return (ret);
 }
 
@@ -1048,13 +1056,12 @@ touch(q)
  * Checks if specified job name is in the printer's queue.
  * Returns:  negative (-1) if argument name is not in the queue.
  */
-static int
-doarg(job)
-	char *job;
+int
+doarg(char *job)
 {
-	register struct queue **qq;
-	register int jobnum, n;
-	register char *cp, *machine;
+	struct queue **qq;
+	int jobnum, fd, n;
+	char *cp, *machine;
 	int cnt = 0;
 	FILE *fp;
 
@@ -1098,15 +1105,18 @@ doarg(job)
 	 * Process item consisting of owner's name (example: henry).
 	 */
 	for (qq = queue + nitems; --qq >= queue; ) {
-		seteuid(euid);
-		fp = fopen((*qq)->q_name, "r");
-		seteuid(uid);
-		if (fp == NULL)
+		PRIV_START;
+		fd = safe_open((*qq)->q_name, O_RDONLY|O_NOFOLLOW, 0);
+		PRIV_END;
+		if (fd < 0 || (fp = fdopen(fd, "r")) == NULL) {
+			if (fd >= 0)
+				close(fd);
 			continue;
-		while (getline(fp) > 0)
+		}
+		while (getcfline(fp) > 0)
 			if (line[0] == 'P')
 				break;
-		(void) fclose(fp);
+		(void)fclose(fp);
 		if (line[0] != 'P' || strcmp(job, line+1) != 0)
 			continue;
 		if (touch(*qq) == 0) {
@@ -1121,25 +1131,23 @@ doarg(job)
  * Enable everything and start printer (undo `down').
  */
 void
-up(argc, argv)
-	int argc;
-	char *argv[];
+up(int argc, char **argv)
 {
-	register int c, status;
-	register char *cp1, *cp2;
+	int c, status;
+	char *cp1, *cp2;
 	char prbuf[100];
 
 	if (argc == 1) {
 		printf("Usage: up {all | printer ...}\n");
 		return;
 	}
-	if (argc == 2 && !strcmp(argv[1], "all")) {
+	if (argc == 2 && strcmp(argv[1], "all") == 0) {
 		printer = prbuf;
 		while (cgetnext(&bp, printcapdb) > 0) {
 			cp1 = prbuf;
 			cp2 = bp;
 			while ((c = *cp2++) && c != '|' && c != ':' &&
-			    (cp1 - prbuf) < sizeof(prbuf))
+			    (cp1 - prbuf) < sizeof(prbuf) - 1)
 				*cp1++ = c;
 			*cp1 = '\0';
 			startpr(2);
