@@ -1,4 +1,4 @@
-/*	$OpenBSD: lpd.c,v 1.40 2003/09/03 20:23:26 tedu Exp $ */
+/*	$OpenBSD: lpd.c,v 1.44 2004/04/07 14:09:35 aaron Exp $ */
 /*	$NetBSD: lpd.c,v 1.33 2002/01/21 14:42:29 wiz Exp $	*/
 
 /*
@@ -41,7 +41,7 @@ static const char copyright[] =
 #if 0
 static const char sccsid[] = "@(#)lpd.c	8.7 (Berkeley) 5/10/95";
 #else
-static const char rcsid[] = "$OpenBSD: lpd.c,v 1.40 2003/09/03 20:23:26 tedu Exp $";
+static const char rcsid[] = "$OpenBSD: lpd.c,v 1.44 2004/04/07 14:09:35 aaron Exp $";
 #endif
 #endif /* not lint */
 
@@ -67,7 +67,7 @@ static const char rcsid[] = "$OpenBSD: lpd.c,v 1.40 2003/09/03 20:23:26 tedu Exp
  *	   and are mode 660.
  *	3. lpd runs as root but spends most of its time with its effective
  *	   uid and gid set to the uid/gid specified in the passwd entry for
- *	   DEFUID (1, aka daemon).
+ *	   DEFUID ("lp").
  *	4. lpr and lprm run setuid daemon and setgrp daemon.  lpr opens
  *	   files to be printed with its real uid/gid and writes to
  *	   the spool dir with its effective uid/gid (i.e. daemon).
@@ -107,6 +107,7 @@ static const char rcsid[] = "$OpenBSD: lpd.c,v 1.40 2003/09/03 20:23:26 tedu Exp
 #include "extern.h"
 
 #define	LPD_NOPORTCHK	0001		/* skip reserved-port check */
+#define _PATH_FHSLOCK	"/var/spool/lpd/lpd.lock"	/* symlink to _PATH_MASTERLOCK to satisfy FHS */
 
 int	lflag;				/* log requests flag */
 int	rflag;				/* allow 'of' for remote printers */
@@ -124,7 +125,7 @@ static void		doit(void);
 static void		startup(void);
 static void		chkhost(struct sockaddr *, int);
 static int		ckqueue(char *);
-static __dead void	usage(void);
+static void		usage(void);
 static int		*socksetup(int, int, const char *);
 
 extern int		__ivaliduser_sa(FILE *, struct sockaddr *, socklen_t,
@@ -144,7 +145,7 @@ main(int argc, char **argv)
 	int lfd, i, f, funix, *finet;
 	int options, check_options, maxfd;
 	long l;
-	long child_max = 32;	/* more then enough to hose the system */
+	long child_max = 32;	/* more than enough to hose the system */
 	struct servent *sp;
 	const char *port = "printer";
 	char *cp;
@@ -155,27 +156,42 @@ main(int argc, char **argv)
 	/*
 	 * We want to run with euid of daemon most of the time.
 	 */
-	if ((pw = getpwuid(DEFUID)) == NULL)
-		errx(1, "daemon uid (%u) not in password file", DEFUID);
+	if ((pw = getpwnam(DEFUID)) == NULL)
+		errx(1, "'lp' uid not in password file");
 	real_uid = pw->pw_uid;
 	real_gid = pw->pw_gid;
 	effective_uid = 0;
 	effective_gid = getegid();
+
+ 	setresgid(real_gid, real_gid, effective_gid);
+	setresuid(real_uid, real_uid, effective_uid);
 	PRIV_END;	/* run as daemon for most things */
 
 	check_options = LPD_NOPORTCHK;		/* XXX - lp* not setuid root */
 	options = 0;
 	gethostname(host, sizeof(host));
 
-	while ((i = getopt(argc, argv, "b:cdln:rsw:W")) != -1) {
+	while ((i = getopt(argc, argv, "b:dln:rsw:W")) != -1) {
 		switch (i) {
 		case 'b':
 			if (blist_addrs >= blist_size) {
-				blist_size += sizeof(char *) * 4;
-				if (blist == NULL)
+				if (blist == NULL) {
+					blist_size += sizeof(char *) * 4;
 					blist = malloc(blist_size);
-				else
-					blist = realloc(blist, blist_size);
+				}
+				else {
+					char **newblist;
+					int newblist_size = blist_size +
+					    sizeof(char *) * 4;
+					newblist = realloc(blist, newblist_size);
+					if (newblist == NULL) {
+						free(blist);
+						blist_size = 0;
+						blist = NULL;
+					}
+					blist = newblist;
+					blist_size = newblist_size;
+				}
 				if (blist == NULL)
 					err(1, "cant allocate bind addr list");
 			}
@@ -252,9 +268,10 @@ main(int argc, char **argv)
 	syslog(LOG_INFO, "restarted");
 	(void)umask(0);
 	PRIV_START;
-	lfd = open(_PATH_MASTERLOCK, O_WRONLY|O_CREAT|O_EXLOCK|O_NONBLOCK, 0644);
+	lfd = open(_PATH_MASTERLOCK, O_WRONLY|O_CREAT/*|O_EXLOCK*/|O_NONBLOCK, 0644);
+	symlink(_PATH_MASTERLOCK, _PATH_FHSLOCK);
 	PRIV_END;
-	if (lfd < 0) {
+	if (lfd < 0 || flock(lfd, LOCK_EX|LOCK_NB) < 0) {
 		if (errno == EWOULDBLOCK)	/* active daemon present */
 			exit(0);
 		syslog(LOG_ERR, "%s: %m", _PATH_MASTERLOCK);
@@ -341,7 +358,7 @@ main(int argc, char **argv)
 	memset(&fromunix, 0, sizeof(fromunix));
 	for (;;) {
 		int domain, nfds, s;
-		socklen_t fromlen;
+		socklen_t fromlen = -1;
 		fd_set readfds;
 		short sleeptime = 10;	/* overflows in about 2 hours */
 
@@ -357,7 +374,9 @@ main(int argc, char **argv)
 			}
 		}
 
-		FD_COPY(&defreadfds, &readfds);
+		/*FD_COPY(defreadfds, readfds);*/
+		memcpy( &readfds, &defreadfds, sizeof(readfds));
+
 		nfds = select(maxfd + 1, &readfds, NULL, NULL, NULL);
 		if (nfds <= 0) {
 			if (nfds < 0 && errno != EINTR)
@@ -453,13 +472,14 @@ reapchild(int signo)
 static void
 mcleanup(int signo)
 {
-	struct syslog_data sdata = SYSLOG_DATA_INIT;
+/*	struct syslog_data sdata = SYSLOG_DATA_INIT; */
 
 	if (lflag)
-		syslog_r(LOG_INFO, &sdata, "exiting");
+		syslog(LOG_INFO, /*&sdata,*/ "exiting");
 	PRIV_START;
 	unlink(_PATH_SOCKETNAME);
 	unlink(_PATH_MASTERLOCK);
+	unlink(_PATH_FHSLOCK);
 	_exit(0);
 }
 
@@ -506,8 +526,8 @@ doit(void)
 			if (*cp >= '\1' && *cp <= '\5') {
 				syslog(LOG_INFO, "%s requests %s %s",
 					from, cmdnames[(int)*cp], cp+1);
-				setproctitle("serving %s: %s %s", from,
-				    cmdnames[(int)*cp], cp+1);
+				/*setproctitle("serving %s: %s %s", from,
+				    cmdnames[(int)*cp], cp+1);*/
 			} else
 				syslog(LOG_INFO, "bad request (%d) from %s",
 					*cp, from);
@@ -628,7 +648,7 @@ startup(void)
 			/* NOTREACHED */
 		case 0:
 			printer = buf;
-			setproctitle("working on printer %s", printer);
+			/*setproctitle("working on printer %s", printer);*/
 			cgetclose();
 			printjob();
 			/* NOTREACHED */
@@ -682,10 +702,16 @@ chkhost(struct sockaddr *f, int check_opts)
 	char host[NI_MAXHOST], ip[NI_MAXHOST];
 	char serv[NI_MAXSERV];
 	int error;
+	int len = 0;
 
-	error = getnameinfo(f, f->sa_len, NULL, 0, serv, sizeof(serv),
+	if( f->sa_family == AF_INET6 )
+	    len = sizeof( struct sockaddr_in6 );
+	else if( f->sa_family == AF_INET )
+	    len = sizeof( struct sockaddr_in );
+
+	error = getnameinfo(f, /*f->sa_*/len, NULL, 0, serv, sizeof(serv),
 	    NI_NUMERICSERV);
-	if (error)
+	if (error || !len)
 		fatal("Malformed from address");
 
 	if (!(check_opts & LPD_NOPORTCHK) &&
@@ -693,10 +719,10 @@ chkhost(struct sockaddr *f, int check_opts)
 		fatal("Connect from invalid port (%s)", serv);
 
 	/* Need real hostname for temporary filenames */
-	error = getnameinfo(f, f->sa_len, host, sizeof(host), NULL, 0,
+	error = getnameinfo(f, /*f->sa_*/len, host, sizeof(host), NULL, 0,
 	    NI_NAMEREQD);
 	if (error) {
-		error = getnameinfo(f, f->sa_len, host, sizeof(host), NULL, 0,
+		error = getnameinfo(f, /*f->sa_*/len, host, sizeof(host), NULL, 0,
 		    NI_NUMERICHOST);
 		if (error)
 			fatal("Host name for your address unknown");
@@ -708,7 +734,7 @@ chkhost(struct sockaddr *f, int check_opts)
 	from = fromb;
 
 	/* need address in stringform for comparison (no DNS lookup here) */
-	error = getnameinfo(f, f->sa_len, host, sizeof(host), NULL, 0,
+	error = getnameinfo(f, /*f->sa_*/len, host, sizeof(host), NULL, 0,
 	    NI_NUMERICHOST);
 	if (error)
 		fatal("Cannot print address");
@@ -732,13 +758,13 @@ chkhost(struct sockaddr *f, int check_opts)
 		freeaddrinfo(res);
 	if (good == 0)
 		fatal("address for your hostname (%s) not matched", host);
-	setproctitle("serving %s", from);
+	/* setproctitle("serving %s", from); */
 	PRIV_START;
 	hostf = fopen(_PATH_HOSTSEQUIV, "r");
 	PRIV_END;
 again:
 	if (hostf) {
-		if (__ivaliduser_sa(hostf, f, f->sa_len, DUMMY, DUMMY) == 0) {
+		if (__ivaliduser_sa(hostf, f, len, DUMMY, DUMMY) == 0) {
 			(void)fclose(hostf);
 			return;
 		}
@@ -755,7 +781,7 @@ again:
 	/*NOTREACHED*/
 }
 
-static __dead void
+static void
 usage(void)
 {
 	extern char *__progname;
@@ -774,7 +800,7 @@ int *
 socksetup(int af, int options, const char *port)
 {
 	struct addrinfo hints, *res, *r;
-	int error, maxs = 0, *s, *socks = NULL, blidx = 0;
+	int error, maxs = 0, *s, *socks = NULL, *newsocks, blidx = 0;
 	const int on = 1;
 
 	do {
@@ -800,8 +826,15 @@ socksetup(int af, int options, const char *port)
 			socks = malloc((maxs + 1) * sizeof(int));
 			if (socks)
 				*socks = 0; /* num of sockets ctr at start */
-		} else
-			socks = realloc(socks, (maxs + 1) * sizeof(int));
+		} else {
+			newsocks = realloc(socks, (maxs + 1) * sizeof(int));
+			if (newsocks)
+				socks = newsocks;
+			else {
+				free(socks);
+				socks = NULL;
+			}
+		}
 		if (!socks) {
 			syslog(LOG_ERR, "couldn't allocate memory for sockets");
 			mcleanup(0);
@@ -809,6 +842,8 @@ socksetup(int af, int options, const char *port)
 
 		s = socks + *socks + 1;
 		for (r = res; r; r = r->ai_next) {
+			int bind_retry = 0;
+			
 			*s = socket(r->ai_family, r->ai_socktype,
 			            r->ai_protocol);
 			if (*s < 0) {
@@ -823,14 +858,32 @@ socksetup(int af, int options, const char *port)
 					close (*s);
 					continue;
 				}
-			PRIV_START;
-			error = bind(*s, r->ai_addr, r->ai_addrlen);
-			PRIV_END;
-			if (error < 0) {
+
+			if( r->ai_family == AF_INET6 ){
+				struct protoent * pe = getprotobyname( "ipv6" );
+				if( pe && setsockopt( *s, IPPROTO_IPV6/*pe->p_proto*/, IPV6_V6ONLY, &on, sizeof(on)) == -1 ){
+					syslog( LOG_DEBUG, "setsockopt() failed: %m - IPV6 socked NOT opened");
+					continue;
+				}
+			}
+
+			while( bind_retry++ < 5 ){
+				PRIV_START;
+				error = bind(*s, r->ai_addr, r->ai_addrlen);
+				PRIV_END;
+				if (error < 0 && errno == EADDRINUSE ){
+					syslog(LOG_DEBUG, "bind(): %m -- Waiting 7 sec. and retrying");
+					sleep(7);
+					continue;
+				}
+				break;
+			}
+			if( error < 0 ){
 				syslog(LOG_DEBUG, "bind(): %m");
 				close (*s);
-				continue;
 			}
+			
+				
 			*socks = *socks + 1;
 			s++;
 		}
