@@ -1,4 +1,4 @@
-/*	$OpenBSD: printjob.c,v 1.37 2003/06/02 23:36:53 millert Exp $	*/
+/*	$OpenBSD: printjob.c,v 1.40 2005/04/28 10:14:26 moritz Exp $	*/
 /*	$NetBSD: printjob.c,v 1.31 2002/01/21 14:42:30 wiz Exp $	*/
 
 /*
@@ -68,6 +68,7 @@ static const char sccsid[] = "@(#)printjob.c	8.7 (Berkeley) 5/10/95";
 #include <stdlib.h>
 #include <stdarg.h>
 #include <ctype.h>
+#include <time.h>
 
 #include "lp.h"
 #include "lp.local.h"
@@ -104,10 +105,10 @@ static int	 tof;		/* true if at top of form */
 static char	class[32];		/* classification field */
 static char	fromhost[MAXHOSTNAMELEN]; /* user's host machine */
 				/* indentation size in static characters */
-static char	indent[10] = "-i0"; 
+static char	indent[10] = "-i0";
 static char	jobname[NAME_MAX];	/* job or file name */
 static char	length[10] = "-l";	/* page length in lines */
-static char	logname[MAXLOGNAME];	/* user's login name */
+static char	logname[LOGIN_NAME_MAX /*MAXLOGNAME*/];	/* user's login name */
 static char	pxlength[10] = "-y";	/* page length in pixels */
 static char	pxwidth[10] = "-x";	/* page width in pixels */
 static char	tempfile[] = "errsXXXXXXXXXX"; /* file name for filter output */
@@ -164,7 +165,7 @@ printjob(void)
 		(void)close(fd);
 	}
 	pid = getpid();				/* for use with lprm */
-	setpgrp(0, pid);
+	setpgid(0, pid); /*setpgrp(0, pid);*/
 
 	/* we add SIGINT to the mask so abortpr() doesn't kill itself */
 	memset(&sa, 0, sizeof(sa));
@@ -185,8 +186,8 @@ printjob(void)
 
 	(void)mktemp(tempfile);			/* safe */
 
-	lfd = safe_open(LO, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXLOCK, 0640);
-	if (lfd < 0) {
+	lfd = safe_open(LO, O_WRONLY|O_CREAT|O_NOFOLLOW/*|O_EXLOCK*/, 0640);
+	if (lfd < 0 || flock(lfd,LOCK_EX|LOCK_NB)<0) {
 		if (errno == EWOULDBLOCK)	/* active daemon present */
 			exit(0);
 		syslog(LOG_ERR, "%s: %s: %m", printer, LO);
@@ -403,7 +404,7 @@ printit(char *file)
 
 	/* pass 1 */
 
-	while (getline(cfp))
+	while (getcfline(cfp))
 		switch (line[0]) {
 		case 'H':
 			strlcpy(fromhost, line+1, sizeof(fromhost));
@@ -504,7 +505,7 @@ printit(char *file)
 
 pass2:
 	fseek(cfp, 0L, 0);
-	while (getline(cfp))
+	while (getcfline(cfp))
 		switch (line[0]) {
 		case 'L':	/* identification line */
 			if (!SH && HL)
@@ -545,7 +546,7 @@ print(int format, char *file)
 	ssize_t nread;
 	struct stat stb;
 	pid_t pid;
-	char *prog, *av[15], buf[BUFSIZ];
+	char *prog, *av[17], buf[BUFSIZ];
 	int fd, status, serrno;
 	int n, fi, fo, p[2], stopped = 0, nofile;
 
@@ -567,7 +568,7 @@ print(int format, char *file)
 		(void)write(ofd, FF, strlen(FF));
 		tof = 1;
 	}
-	if (IF == NULL && (format == 'f' || format == 'l')) {
+	if (IF == NULL && (format == 'f' || format == 'l' || format == 'o')) {
 		tof = 0;
 		while ((n = read(fi, buf, BUFSIZ)) > 0)
 			if (write(ofd, buf, n) != n) {
@@ -586,7 +587,8 @@ print(int format, char *file)
 			av[2] = length;
 			av[3] = "-h";
 			av[4] = *title ? title : " ";
-			av[5] = NULL;
+			av[5] = "-f";
+			av[6] = NULL;
 			fo = ofd;
 			goto start;
 		}
@@ -598,7 +600,7 @@ print(int format, char *file)
 			nofile = sysconf(_SC_OPEN_MAX);
 			for (n = 3; n < nofile; n++)
 				(void)close(n);
-			execl(_PATH_PR, "pr", width, length,
+			execl(_PATH_PR, "pr", width, length, "-f",
 			    "-h", *title ? title : " ", (char *)NULL);
 			syslog(LOG_ERR, "cannot execl %s", _PATH_PR);
 			exit(2);
@@ -611,14 +613,6 @@ print(int format, char *file)
 			return(ERROR);
 		}
 		fi = p[0];			/* use pipe for input */
-	case 'o':       /* print postscript file */
-		/*
-		 * For now, treat this as a plain-text file, and assume
-		 * the standard LPF_INPUT filter will recognize that it
-		 * is postscript and know what to do with it.  These
-		 * 'o'-file requests could come from MacOS X systems.
-		 */
-		/* FALLTHROUGH */
 	case 'f':	/* print plain text file */
 		prog = IF;
 		av[1] = width;
@@ -626,6 +620,17 @@ print(int format, char *file)
 		av[3] = indent;
 		n = 4;
 		break;
+	case 'o':       /* print postscript file */
+		/*
+		 * Treat this as a "plain file with control characters", and
+		 * assume the standard LPF_INPUT filter will recognize that
+		 * the data is postscript and know what to do with it.  These
+		 * 'o'-file requests could come from MacOS 10.1 systems.
+		 * (later versions of MacOS 10 will explicitly use 'l')
+		 * A postscript file can contain binary data, which is why 'l'
+		 * is somewhat more appropriate than 'f'.
+		 */
+		/* FALLTHROUGH */
 	case 'l':	/* like 'f' but pass control characters */
 		prog = IF;
 		av[1] = "-c";
@@ -699,6 +704,10 @@ print(int format, char *file)
 		av[0] = prog;
 	av[n++] = "-n";
 	av[n++] = logname;
+	if (*jobname != '\0' && strcmp(jobname, " ") != 0) {
+		av[n++] = "-j";
+		av[n++] = jobname;
+	}
 	av[n++] = "-h";
 	av[n++] = fromhost;
 	av[n++] = AF;
@@ -712,8 +721,8 @@ print(int format, char *file)
 		if (WIFSTOPPED(status) == 0) {
 			(void)close(fi);
 			syslog(LOG_WARNING,
-			    "%s: output filter died (retcode=%d termsig=%d)",
-			    printer, WEXITSTATUS(status), WTERMSIG(status));
+			    "%s: output filter died (retcode=%d termsig=%d prog=%s)",
+			    printer, WEXITSTATUS(status), WTERMSIG(status), prog);
 			return(REPRINT);
 		}
 		stopped++;
@@ -763,8 +772,8 @@ start:
 	}
 
 	if (!WIFEXITED(status)) {
-		syslog(LOG_WARNING, "%s: filter '%c' terminated (termsig=%d)",
-		    printer, format, WTERMSIG(status));
+		syslog(LOG_WARNING, "%s: filter '%c' terminated (termsig=%d prog=%s)",
+		    printer, format, WTERMSIG(status), prog);
 		return(ERROR);
 	}
 	switch (WEXITSTATUS(status)) {
@@ -776,8 +785,8 @@ start:
 	case 2:
 		return(ERROR);
 	default:
-		syslog(LOG_WARNING, "%s: filter '%c' exited (retcode=%d)",
-		    printer, format, WEXITSTATUS(status));
+		syslog(LOG_WARNING, "%s: filter '%c' exited (retcode=%d prog=%s)",
+		    printer, format, WEXITSTATUS(status), prog);
 		return(FILTERERR);
 	}
 }
@@ -812,7 +821,7 @@ sendit(char *file)
 	/*
 	 * pass 1
 	 */
-	while (getline(cfp)) {
+	while (getcfline(cfp)) {
 	again:
 		if (line[0] == 'S') {
 			cp = line+1;
@@ -829,7 +838,7 @@ sendit(char *file)
 		}
 		if (line[0] >= 'a' && line[0] <= 'z') {
 			strlcpy(last, line, sizeof(last));
-			while ((i = getline(cfp)) != 0)
+			while ((i = getcfline(cfp)) != 0)
 				if (strcmp(last, line))
 					break;
 			switch (sendfile('\3', last+1)) {
@@ -856,7 +865,7 @@ sendit(char *file)
 	 * pass 2
 	 */
 	fseek(cfp, 0L, 0);
-	while (getline(cfp))
+	while (getcfline(cfp))
 		if (line[0] == 'U' && strchr(line+1, '/') == 0)
 			(void)unlink(line+1);
 	/*
@@ -996,21 +1005,21 @@ banner(char *name1, char *name2)
 		(void)write(ofd, name2, strlen(name2));
 		(void)write(ofd, "  Date: ", 8);
 		(void)write(ofd, ctime(&tvec), 24);
-		(void)write(ofd, "\n", 1);
+		(void)write(ofd, "\n\r", 2);
 	} else {	/* normal banner */
-		(void)write(ofd, "\n\n\n", 3);
+		(void)write(ofd, "\n\r\n\r\n\r", 6);
 		scan_out(ofd, name1, '\0');
-		(void)write(ofd, "\n\n", 2);
+		(void)write(ofd, "\n\r\n\r", 4);
 		scan_out(ofd, name2, '\0');
 		if (class[0]) {
-			(void)write(ofd,"\n\n\n",3);
+			(void)write(ofd,"\n\r\n\r\n\r",6);
 			scan_out(ofd, class, '\0');
 		}
-		(void)write(ofd, "\n\n\n\n\t\t\t\t\tJob:  ", 15);
+		(void)write(ofd, "\n\r\n\r\n\r\n\r\t\t\t\t\tJob:  ", 19);
 		(void)write(ofd, name2, strlen(name2));
-		(void)write(ofd, "\n\t\t\t\t\tDate: ", 12);
+		(void)write(ofd, "\n\r\t\t\t\t\tDate: ", 13);
 		(void)write(ofd, ctime(&tvec), 24);
-		(void)write(ofd, "\n", 1);
+		(void)write(ofd, "\n\r", 2);
 	}
 	if (!SF)
 		(void)write(ofd, FF, strlen(FF));
@@ -1036,7 +1045,7 @@ scan_out(int scfd, char *scsp, int dlm)
 {
 	char *strp;
 	int nchrs, j;
-	char outbuf[LINELEN+1], *sp, c, cc;
+	char outbuf[LINELEN+2], *sp, c, cc;
 	int d, scnhgt;
 	extern char scnkey[][HEIGHT];	/* in lpdchar.c */
 
@@ -1060,7 +1069,8 @@ scan_out(int scfd, char *scsp, int dlm)
 		while (*--strp == BACKGND && strp >= outbuf)
 			;
 		strp++;
-		*strp++ = '\n';	
+		*strp++ = '\n';
+		*strp++ = '\r';
 		(void)write(scfd, outbuf, strp-outbuf);
 	}
 }
@@ -1195,7 +1205,6 @@ dofork(int action)
 			}
 			initgroups(pw->pw_name, pw->pw_gid);
 			setgid(pw->pw_gid);
-			setlogin("");
 			setuid(DU);
 		}
 		return (pid);
@@ -1266,7 +1275,7 @@ init(void)
 	if (cgetstr(bp, "sd", &SD) == -1)
 		SD = _PATH_DEFSPOOL;
 	if (cgetnum(bp, "du", &DU) < 0)
-		DU = DEFUID;
+		DU = DEFNUID;
 	if (cgetstr(bp,"ff", &FF) == -1)
 		FF = DEFFF;
 	if (cgetnum(bp, "pw", &PW) < 0)
@@ -1491,24 +1500,24 @@ struct bauds {
 	int	baud;
 	int	speed;
 } bauds[] = {
-	50,	B50,
-	75,	B75,
-	110,	B110,
-	134,	B134,
-	150,	B150,
-	200,	B200,
-	300,	B300,
-	600,	B600,
-	1200,	B1200,
-	1800,	B1800,
-	2400,	B2400,
-	4800,	B4800,
-	9600,	B9600,
-	19200,	B19200,
-	38400,	B38400,
-	57600,	B57600,
-	115200,	B115200,
-	0,	0
+	{50,	B50},
+	{75,	B75},
+	{110,	B110},
+	{134,	B134},
+	{150,	B150},
+	{200,	B200},
+	{300,	B300},
+	{600,	B600},
+	{1200,	B1200},
+	{1800,	B1800},
+	{2400,	B2400},
+	{4800,	B4800},
+	{9600,	B9600},
+	{19200,	B19200},
+	{38400,	B38400},
+	{57600,	B57600},
+	{115200,	B115200},
+	{0,	0}
 };
 #endif
 
@@ -1540,7 +1549,7 @@ setty(void)
 			if (BR == bp->baud)
 				break;
 		if (!bp->baud) {
-			syslog(LOG_ERR, "%s: illegal baud rate %d", printer, BR);
+			syslog(LOG_ERR, "%s: illegal baud rate %d", printer, (int)BR);
 			exit(1);
 		}
 		cfsetspeed(&i.t, bp->speed);
@@ -1587,19 +1596,19 @@ setty(void)
 		}
 	} else {
 		if (FC) {
-			sttyclearflags(&i.t, FC);
+			i.t.c_cflag &= ~FC;
 			i.set = 1;
 		}
 		if (FS) {
-			sttysetflags(&i.t, FS);
+			i.t.c_cflag |= FS;
 			i.set = 1;
 		}
 		if (XC) {
-			sttyclearlflags(&i.t, XC);
+			i.t.c_lflag &= ~XC;
 			i.set = 1;
 		}
 		if (XS) {
-			sttysetlflags(&i.t, XS);
+			i.t.c_lflag |= XS;
 			i.set = 1;
 		}
 	}
@@ -1622,14 +1631,16 @@ pstatus(const char *msg, ...)
 
 	va_start(ap, msg);
 	umask(0);
-	fd = open(ST, O_WRONLY|O_CREAT|O_NOFOLLOW|O_EXLOCK, 0660);
-	if (fd < 0) {
+	fd = open(ST, O_WRONLY|O_CREAT|O_NOFOLLOW /*|O_EXLOCK*/, 0660);
+	if (fd < 0 || flock(fd,LOCK_EX|LOCK_NB)<0) {
 		syslog(LOG_ERR, "%s: %s: %m", printer, ST);
 		exit(1);
 	}
 	ftruncate(fd, 0);
 	len = vsnprintf(buf, sizeof(buf), msg, ap);
 	va_end(ap);
+	if (len < 0)			/* Output error */
+		return;
 	if (len >= sizeof(buf))
 		len = sizeof(buf) - 1;
 	buf[len++] = '\n';		/* replace NUL with newline */
