@@ -117,11 +117,8 @@ main(int argc, char **argv)
 		fatal("Your name is too long");
 	strlcpy(luser, pw->pw_name, sizeof(luser));
 	person = luser;
-	while ((ch = getopt(argc, argv, "P:w:-")) != -1) {
+	while ((ch = getopt(argc, argv, "P:w")) != -1) {
 		switch (ch) {
-		case '-':
-			users = -1;
-			break;
 		case 'P':
 			printer = optarg;
 			break;
@@ -143,6 +140,12 @@ main(int argc, char **argv)
 
 	if (printer == NULL && (printer = getenv("PRINTER")) == NULL)
 		printer = DEFLP;
+	/* getopt doesn't handle '-' as an argument so we have to handle it here manually */
+	if(argc == 1 && *argv[0] == '-' && strlen(argv[0]) == 1) {
+		argc--;
+		argv++;
+		users = -1;
+	}
 	if (users < 0 && argc != 0)
 		usage();
 	while (argc > 0) {
