@@ -63,6 +63,7 @@ static const char rcsid[] = "$OpenBSD: pac.c,v 1.17 2003/06/02 23:36:53 millert
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <signal.h>
 
 #include "lp.h"
 #include "lp.local.h"
@@ -107,7 +108,7 @@ static struct	hent *enter(const char *);
 static struct	hent *lookup(const char *);
 static int	qucmp(const void *, const void *);
 static void	rewrite(void);
-__dead void	usage(void);
+void	usage(void);
 
 int
 main(int argc, char **argv)
@@ -465,7 +466,7 @@ chkprinter(const char *s)
 	return(1);
 }
 
-__dead void
+void
 usage(void)
 {
 	extern char *__progname;
